from typing import List, Dict
import numpy as np
import os

from utils import misc_utils, table_utils
from trajectory.trajectory import Trajectory
import sklearn.metrics as metrics
import matplotlib.pyplot as plt
from multiprocessing import Pool


def run_methods_on_datasets(fnames: List[str], result_dirs: List[str], methods: List[str], num_procs: int):
    # TODO incorporate start and end periods
    # get commands for running data
    cmds: List[str] = []
    for dataset_fname, result_dir in zip(fnames, result_dirs):
        for method in methods:
            method = method.upper()

            result_dir_method: str = f"{result_dir}/{method}/"
            misc_utils.makedirs_if_needed(result_dir_method)
            if method == "BC":
                cmd = f"python BioCycle.py --input_file {dataset_fname} --output_dir {result_dir_method}"
            elif method == "BCO":
                cmd: str = f"Rscript ../old/BioCycle/BioCycle.R -i {dataset_fname} -o {result_dir_method}"
            elif method in ["ARS", "LS", "JTK"]:
                if method == "ARS":
                    # average replicates
                    tps, _, _, _ = table_utils.get_timepoints(dataset_fname)
                    trajs: List[Trajectory] = table_utils.get_trajs(dataset_fname)

                    header = ["ID"] + [str(x) for x in tps]
                    rows: List[List[str]] = []

                    for traj in trajs:
                        row: List[str] = [traj.name]
                        for tp in tps:
                            if tp in traj.tps:
                                tp_idx = traj.tps.index(tp)
                                val: str = str(np.mean(traj.reps[tp_idx]))
                            else:
                                val: str = "NA"

                            row.append(val)
                        rows.append(row)

                    dataset_fname_ave: str = f"{result_dir_method}/input_data_ave.tsv"
                    table_utils.write_table(dataset_fname_ave, header, rows)

                    cmd = f"python MetaCycle/MetaCycle.py --input_file {dataset_fname_ave} --output_dir " \
                          f"{result_dir_method} --methods {method}"
                else:
                    cmd = f"python MetaCycle/MetaCycle.py --input_file {dataset_fname} --output_dir " \
                          f"{result_dir_method} --methods {method}"
            else:
                raise ValueError("Unknown method %s" % method)

            print(cmd)
            cmds.append(cmd)

    # run commands
    pool = Pool(num_procs)
    pool.map(os.system, cmds)
    pool.close()
    pool.join()


def plot_roc_auc(true_labels: np.array, method_preds: List[np.array], methods: List[str], plot_fname: str):
    color_dir: Dict[str, str] = {"BC": "black", "JTK": "red", "ARS": "blue", "LS": "green"}
    for method_pred, method in zip(method_preds, methods):
        method = method.upper()
        color = color_dir[method]

        fpr, tpr, _ = metrics.roc_curve(true_labels, method_pred)
        roc_auc = metrics.auc(fpr, tpr)
        plt.plot(fpr, tpr, 'b', label='%s AUC = %0.2f' % (method, roc_auc), color=color)

        print("%s: %f" % (method, roc_auc))

    plt.title('Receiver Operating Characteristic')
    plt.legend(loc='lower right')
    plt.plot([0, 1], [0, 1], 'r--')
    plt.xlim([0, 1])
    plt.ylim([0, 1])
    plt.ylabel('True Positive Rate')
    plt.xlabel('False Positive Rate')
    plt.savefig(plot_fname)
